/* translatable strings for gretl addons */

const char *gig_translations[] = {
    N_("gig"),
    N_("APARCH"),
    N_("AR lags"),
    N_("ARCH"),
    N_("Constant"),
    N_("Covariance estimator"),
    N_("Density"),
    N_("Dependent Variable"),
    N_("Distribution"),
    N_("EGARCH"),
    N_("Forecast"),
    N_("GARCH"),
    N_("GED"),
    N_("GJR"),
    N_("Hessian"),
    N_("Mean regressors"),
    N_("Model type"),
    N_("NARCH"),
    N_("Normal"),
    N_("OPG"),
    N_("Plot type"),
    N_("Sandwich"),
    N_("Skewed GED"),
    N_("Skewed t"),
    N_("TARCH"),
    N_("Taylor/Schwert GARCH"),
    N_("Time series"),
    N_("Variance regressors"),
    N_("Verbosity"),
    N_("t")
};

const char *ivpanel_translations[] = {
    N_("Between model"),
    N_("Fixed effects"),
    N_("G2SLS"),
    N_("dependent variable"),
    N_("instruments"),
    N_("regressors")
};

const char *SVAR_translations[] = {
    N_("Structural VARs"),
    N_("AB-model"),
    N_("BFGS (analytical score)"),
    N_("BFGS (numerical score)"),
    N_("Bias correction"),
    N_("Bootstrap alpha"),
    N_("Bootstrap replications"),
    N_("Bootstrap type"),
    N_("C-model"),
    N_("CI coverage"),
    N_("Check identification"),
    N_("Constant"),
    N_("Exogenous regressors"),
    N_("FEVD"),
    N_("Full"),
    N_("Historical decomposition"),
    N_("Horizon"),
    N_("IRF"),
    N_("Indices of responses to cumulate"),
    N_("Lags"),
    N_("Model type"),
    N_("Newton-Raphson (analytical score)"),
    N_("Newton-Raphson (numerical score)"),
    N_("None"),
    N_("Optimization method"),
    N_("Partial"),
    N_("Plot type"),
    N_("Restriction pattern (long-run C or A)"),
    N_("Restriction pattern (short-run C or B)"),
    N_("Scoring algorithm"),
    N_("Seasonal dummies"),
    N_("Time trend"),
    N_("VAR variables"),
    N_("bias correction (non-SVEC)"),
    N_("bootstrap iterations"),
    N_("bootstrap type"),
    N_("choose 1 for medians"),
    N_("mean"),
    N_("measure to target"),
    N_("median"),
    N_("moving blocks"),
    N_("plain (Cholesky)"),
    N_("residual resampling"),
    N_("the model bundle"),
    N_("which shocks"),
    N_("which variables"),
    N_("wild/Mammen"),
    N_("wild/Normal"),
    N_("wild/Rademacher")
};

const char *HIP_translations[] = {
    N_("IV/Heteroskedastic"),
    N_("Covariance matrix estimation"),
    N_("Dependent variable"),
    N_("Endogenous variables"),
    N_("Exogenous variables"),
    N_("Hessian"),
    N_("Instruments"),
    N_("OPG"),
    N_("Sandwich"),
    N_("Variance regressors"),
    N_("Verbosity level")
};

const char *dbnomics_translations[] = {
    N_("dbnomics"),
};

const char *extra_translations[] = {
    N_("Don't"),
    N_("Fill method"),
    N_("Linear interpolation"),
    N_("Mean value"),
    N_("Number of random variables"),
    N_("Number of selected choices"),
    N_("Repetition"),
    N_("Series to fill"),
    N_("Set of all choices"),
    N_("Std. deviation"),
    N_("Vector of values"),
    N_("W test stat"),
    N_("always use normal approx"),
    N_("bootstrap type code"),
    N_("left truncation value"),
    N_("moving blocks"),
    N_("number of trials"),
    N_("options"),
    N_("resampling"),
    N_("right truncation value"),
    N_("two-sided prob mass"),
    N_("wildM"),
    N_("wildN"),
    N_("wildR")
};

const char *regls_translations[] = {
    N_("Coefficient paths"),
    N_("Criterion plot")
};

const char *KFgui_translations[] = {
    N_("Absolute value"),
    N_("Disturbance in obs. equation?"),
    N_("Exponential"),
    N_("Normal"),
    N_("Observation matrix (Z)"),
    N_("Parametrization"),
    N_("Quiet"),
    N_("Square"),
    N_("State transition matrix (T)"),
    N_("State variance factor matrix (R)"),
    N_("Verbose"),
    N_("do smoothing"),
    N_("observables"),
    N_("verbosity")
};

